#include "StdAfx.h"
#include "SRAF.h"
#include "Image.h"
#include "EdgeDetector.h"
#include "Canny.h"
#include <math.h>

#define PI 3.14159265359f

/* Scale floating point magnitudes and angles to 8 bits */
#define ORI_SCALE 40.0
#define MAG_SCALE 20.0

/* Biggest possible filter mask */
#define MAX_MASK_SIZE 20

/* Fraction of pixels that should be above the HIGH threshold */
float ratio = 0.1f;
int WIDTH = 0;

IMPLEMENT_DYNAMIC(CCanny, CEdgeDetector);

CCanny::CCanny()
{
	m_low = 5;
	m_high = 3;
	m_sigma = 1.0;
}

CCanny::CCanny(CCanny& c)
{
	m_low = c.m_low;
	m_high = c.m_high;
	m_sigma = c.m_sigma;
}

CCanny::~CCanny()
{
}

void CCanny::GetParams(CString& params)
{
	params.Format("Thresh. bajo: %d, thresh. alto: %d y sigma: &d",
		m_low, m_high, m_sigma);
}

bool CCanny::Detect(CImage& im)
{
	int i, j;

	SetRows(im.GetRows());
	SetCols(im.GetCols());

	if (!m_magim.Create(GetRows(), GetCols()) ||
		!m_oriim.Create(GetRows(), GetCols())) {
		return false;
	}

	/* Apply the filter */
	SetProgress(0, "Detector de contorno Canny");
	
	if (!Canny(im))
		return false;
    
	/* Hysteresis thresholding of edge pixels */
	if (SetProgress(85, "Umbralado hysteresis..."))
		return false;
	
	if (!Hysteresis(im))
		return false;

	for (i = 0; i < WIDTH; i++)
	  for (j = 0; j < GetCols(); j++)
	    im[i][j] = 255;

	for (i = GetRows() - 1; i > GetRows() - 1 - WIDTH; i--)
	  for (j = 0; j < GetCols(); j++)
	    im[i][j] = 255;

	for (i = 0; i < GetRows(); i++)
	  for (j = 0; j < WIDTH; j++)
	    im[i][j] = 255;

	for (i = 0; i < GetRows(); i++)
	  for (j = GetCols() - WIDTH - 1; j < GetCols(); j++)
	    im[i][j] = 255;

	SetProgress(100, "Deteccin de contorno finalizada");

	return true;
}

float CCanny::Norm(float x, float y)
{
	return (float) sqrt ( (double)(x*x + y*y) );
}

bool CCanny::Canny(CImage& img)
{
	int width;
	float **smx,**smy;
	float **dx,**dy;
	int i, j, n;
	float gau[MAX_MASK_SIZE], dgau[MAX_MASK_SIZE], z;

/* Create a Gaussian and a derivative of Gaussian filter mask */
	for(i = 0; i < MAX_MASK_SIZE; i++)
	{
	  gau[i] = MeanGauss ((float)i);
	  if (gau[i] < 0.005)
	  {
		width = i;
		break;
	  }
	  dgau[i] = DGauss ((float)i);
	}

	n = width + width + 1;
	WIDTH = width / 2;

	if (SetProgress(20, "Suavizando con filtro gaussiano..."))
		return false;

	smx = f2d();
	smy = f2d();

	if(SetProgress(40, "Contorcionando en direcciones X e Y"))
		return false;

	/* Convolution of source image with a Gaussian in X and Y directions  */
	SeperableConvolution(img, gau, width, smx, smy);

	/* Now convolve smoothed data with a derivative */
	if(SetProgress(60, "Contorcionando con la derivada"))
		return false;
	
	dx = f2d();
	DxySeperableConvolution (smx, dgau, width, dx, 1);
	
	Free_f2d(smx);

	dy = f2d();
	DxySeperableConvolution (smy, dgau, width, dy, 0);
	
	Free_f2d(smy);

	/* Create an image of the norm of dx,dy */
	for (i = 0; i < GetRows(); i++)
	  for (j=0; j < GetCols(); j++)
	  {
	      z = Norm (dx[i][j], dy[i][j]);
	      m_magim[i][j] = (unsigned char)(z * MAG_SCALE);
	  }

	/* Non-maximum suppression - edge pixels should be a local max */
	if(SetProgress(75, "Suprimiendo no mximos..."))
		return false;

	NonMaxSuppress (dx, dy);

	Free_f2d(dx);
	Free_f2d(dy);

	return true;
}

/*      Gaussian        */
float CCanny::Gauss(float x)
{
    float xx;

    if (m_sigma == 0) return 0.0;
    xx = (float) exp((double) ((-x * x) / (2 * m_sigma * m_sigma)));
    return xx;
}

float CCanny::MeanGauss(float x)
{
	float z;

	z = (Gauss(x) + Gauss(x + 0.5f) + Gauss(x - 0.5f)) / 3.0f;
	z = z / (PI * 2.0f * m_sigma * m_sigma);
	return z;
}

/*      First derivative of Gaussian    */
float CCanny::DGauss (float x)
{
	return -x / (m_sigma * m_sigma) * Gauss(x);
}

/*      HYSTERESIS thersholding of edge pixels. Starting at pixels with a
	value greater than the HIGH threshold, trace a connected sequence
	of pixels that have a value greater than the LOW threhsold.        */

bool CCanny::Hysteresis (CImage& img)
{
	int i,j;
	int high_thresh = m_high, low_thresh = m_low;

	TRACE("Beginning hysteresis thresholding...\n");
	
	for (i = 0; i < GetRows(); i++)
	  for (j = 0; j < GetCols(); j++)
	    img[i][j] = 0;

	if (high_thresh < low_thresh)
	{
	  EstimateThresh(&low_thresh, &high_thresh);
	  TRACE("Hysteresis thresholds (from image): HI %d LOW %d\n", high_thresh, low_thresh);
	}

/* For each edge with a magnitude above the high threshold, begin
   tracing edge pixels that are above the low threshold.                */

	for (i = 0; i < GetRows(); i++)
	  for (j = 0; j < GetCols(); j++)
	    if (m_magim[i][j] >= high_thresh)
	      Trace(img, i, j, low_thresh);

/* Make the edge black (to be the same as the other methods) */
	for (i = 0; i < GetRows(); i++)
	  for (j = 0; j < GetCols(); j++)
	    if (img[i][j] == 0) img[i][j] = 255;
	    else img[i][j] = 0;
	
	return true;
}

/*      TRACE - recursively trace edge pixels that have a threshold > the low
	edge threshold, continuing from the pixel at (i,j).                     */

int CCanny::Trace(CImage& img, int i, int j, int low)
{
	int n, m;
	char flag = 0;

	if (img[i][j] == 0)
	{
	  img[i][j] = 255;
	  
	  flag = 0;

	  for (n = -1; n <= 1; n++)
	  {
	    for(m = -1; m <= 1; m++)
	    {
	      if (i == 0 && m == 0) continue;
		  if (m_magim.TestRange(i+n,j+m) && 
			  m_magim[i+n][j+m] >= low) 
		  {
			if (Trace(img, i+n, j+m, low))
			{
				flag=1;
				break;
			}
		  }
	    }
	    if (flag) break;
	  }
	  return(1);
	}
	return(0);
}

void CCanny::SeperableConvolution(CImage& img, float *gau, int width, 
								   float **smx, float **smy)
{
	int i,j,k, I1, I2, nr, nc;
	float x, y;

	nr = GetRows();
	nc = GetCols();

	for (i = 0; i < nr; i++) {
	  for (j = 0; j < nc; j++)
	  {
	    x = gau[0] * img[i][j]; 
		y = gau[0] * img[i][j];
	    for (k = 1; k < width; k++)
	    {
	      I1 = (i + k) % nr; 
		  I2 = (i - k + nr) % nr;
	      y += gau[k] * img[I1][j] + gau[k] * img[I2][j];
	      I1 = (j + k) % nc; 
		  I2 = (j - k + nc) % nc;
	      x += gau[k] * img[i][I1] + gau[k] * img[i][I2];
	    }
	    smx[i][j] = x; 
		smy[i][j] = y;
	  }
	}
}

void CCanny::DxySeperableConvolution (float** im, float *gau, 
		int width, float **sm, int which)
{
	int i,j,k, I1, I2;
	float x;

	for (i = 0; i < GetRows(); i++) {
	  for (j = 0; j < GetCols(); j++)
	  {
	    x = 0.0;
	    for (k = 1; k < width; k++)
	    {
	      if (which == 0)
	      {
			I1 = (i + k) % GetRows();
			I2 = (i - k + GetRows()) % GetRows();
			x += -gau[k] * im[I1][j] + gau[k] * im[I2][j];
	      }
	      else
	      {
			I1 = (j + k) % GetCols();
			I2 = (j - k + GetCols()) % GetCols();
			x += -gau[k] * im[i][I1] + gau[k] * im[i][I2];
	      }
	    }
	    sm[i][j] = x;
	  }
	}
}

void CCanny::NonMaxSuppress (float **dx, float **dy)
{
	int i,j;
	float xx, yy, g2, g1, g3, g4, g, xc, yc;

	for (i = 1; i < GetRows() - 1; i++)
	{
	  for (j = 1; j < GetCols() - 1; j++)
	  {
	    m_magim[i][j] = 0;

/* Treat the x and y derivatives as components of a vector */
	    xc = dx[i][j];
	    yc = dy[i][j];
	    if (fabs(xc)<0.01 && fabs(yc)<0.01) continue;

	    g  = Norm(xc, yc);

/* Follow the gradient direction, as indicated by the direction of
   the vector (xc, yc); retain pixels that are a local maximum. */

	    if (fabs(yc) > fabs(xc))
	    {

/* The Y component is biggest, so gradient direction is basically UP/DOWN */
	      xx = (float) (fabs(xc) / fabs(yc));
	      yy = 1.0;

	      g2 = Norm(dx[i-1][j], dy[i-1][j]);
	      g4 = Norm(dx[i+1][j], dy[i+1][j]);
	      if (xc * yc > 0.0) {
			g3 = Norm(dx[i+1][j+1], dy[i+1][j+1]);
			g1 = Norm(dx[i-1][j-1], dy[i-1][j-1]);
	      } 
		  else {
			g3 = Norm(dx[i+1][j-1], dy[i+1][j-1]);
			g1 = Norm(dx[i-1][j+1], dy[i-1][j+1]);
	      }
	    } 
		else {

/* The X component is biggest, so gradient direction is basically LEFT/RIGHT */
	      xx = (float) (fabs(yc) / fabs(xc));
	      yy = 1.0;

	      g2 = Norm(dx[i][j+1], dy[i][j+1]);
	      g4 = Norm(dx[i][j-1], dy[i][j-1]);
	      if (xc * yc > 0.0) {
			g3 = Norm(dx[i-1][j-1], dy[i-1][j-1]);
			g1 = Norm(dx[i+1][j+1], dy[i+1][j+1]);
	      }
	      else {
			g1 = Norm(dx[i-1][j+1], dy[i-1][j+1]);
			g3 = Norm(dx[i+1][j-1], dy[i+1][j-1]);
	      }
	    }

/* Compute the interpolated value of the gradient magnitude */
	    if ( (g > (xx * g1 + (yy - xx) * g2)) &&
		 (g > (xx * g3 + (yy - xx) * g4)) )
	    {
			if (g * MAG_SCALE <= 255)
				m_magim[i][j] = (unsigned char)(g * MAG_SCALE);
			else
				m_magim[i][j] = 255;
	      
			m_oriim[i][j] = (BYTE) (atan2 (yc, xc) * ORI_SCALE);
	    } 
		else {
			m_magim[i][j] = 0;
			m_oriim[i][j] = 0;
	    }
	  }
	}
}

void CCanny::EstimateThresh(int* low_thresh, int* high_thresh)
{
	int i, j, k, hist[256], count;

/* Build a histogram of the magnitude image. */
	for (k = 0; k < 256; k++) hist[k] = 0;

	for (i = WIDTH; i < GetRows() - WIDTH; i++)
	  for (j = WIDTH; j < GetCols() - WIDTH; j++)
	    hist[m_magim[i][j]]++;

/* The high threshold should be > 80 or 90% of the pixels 
	j = (int)(ratio*mag->info->nr*mag->info->nc);
*/
	j = GetRows();
	if (j < GetCols()) j = GetCols();
	j = (int)(0.9 * j);
	k = 255;

	count = hist[255];
	while (count < j)
	{
	  k--;
	  if (k<0) break;
	  count += hist[k];
	}
	*high_thresh = k;

	i=0;
	while (hist[i] == 0) i++;

	*low_thresh = (int) ((*high_thresh + i) / 2.0);
}

#ifdef _DEBUG
void CCanny::Dump( CDumpContext& dc ) const
{
    CObject::Dump( dc );

    dc << "GetRows(): " << GetRows() << "\n"
        << "GetCols(): " << GetCols() << "\n";
}
#endif
