#include "StdAfx.h"
#include "SRAF.h"
#include <math.h>
#include "Image.h"
#include "EdgeDetector.h"
#include "Shen.h"

#define OUTLINE 25

IMPLEMENT_DYNAMIC(CShen, CEdgeDetector);

CShen::CShen()
{
	m_b = 0.035;
	m_low_thresh = 20;
	m_high_thresh = 22;
	m_ratio = 0.99;
	m_window_size = 7;
	m_do_hysteresis = 1;
	m_thinFactor = 1;
	m_lap = NULL;
	m_res = NULL;
}

CShen::CShen(CShen& s)
{
	m_b = s.m_b;
	m_low_thresh = s.m_low_thresh;
	m_high_thresh = s.m_high_thresh;
	m_ratio = s.m_ratio;
	m_window_size = s.m_window_size;
	m_do_hysteresis = s.m_do_hysteresis;
	m_thinFactor = s.m_thinFactor;
	m_lap = NULL;
	m_res = NULL;
}

CShen::~CShen()
{
}

void CShen::GetParams(CString& params)
{
	params.Format("Thresh. bajo: %d, thresh. alto: %d",
		m_low_thresh, m_high_thresh);
}

bool CShen::Detect(CImage& src)
{
	int row, col, i, j, dibRows, dibCols; 
	int width = OUTLINE + 2;

	dibRows = src.GetRows();
	dibCols = src.GetCols();

	SetRows(dibRows + 2 * width);
	SetCols(dibCols + 2 * width);

	SetProgress(0, "Detector de contorno Shen-Castan (ISEF)");

	m_im.Create(GetRows(), GetCols());
	m_res.Create(GetRows(), GetCols());

	/* embed the src in a larger buffer */
	for (row = 0; row < GetRows(); row++) {
		for (col = 0; col < GetCols(); col++) {

			i = (row - width + dibRows) % dibRows;
			j = (col - width + dibCols) % dibCols;
			m_im[row][col] = src[i][j];
		}
	}

	if(SetProgress(15, "Comenzando la deteccin..."))
		return false;

	if (!Shen())
		return false;

	/* copy the result in the src */
	for (row = 0; row < dibRows; row++) {
		for (col = 1; col < dibCols; col++) {

			src[row][col] = m_res[row + width][col + width];
			src[dibRows - 1][col] = 255;
		}
		src[row][0] = 255;
	}
	
	if (SetProgress(100, "Contorno detectado"))
		return false;

	return true;
}

bool CShen::Shen ()
{
	register int i,j;
	float **buffer;	
	float **smoothed_buffer;
	CImage bli_buffer(GetRows(), GetCols());
   
/* Convert the input image to floating point */
	buffer = f2d();
	for (i = 0; i < GetRows(); i++)
	  for (j = 0; j < GetCols(); j++)
	    buffer[i][j] = (float)(m_im[i][j]);

	if(SetProgress(20, "Calculando ISEF...")) return false;
/* Smooth input image using recursively implemented ISEF filter */
	smoothed_buffer = f2d();
	ComputeISEF(buffer, smoothed_buffer);
      
	if(SetProgress(60, "Calculando binivel...")) return false;
/* Compute bli image band-limited laplacian image from smoothed image */
	ComputeBli(smoothed_buffer, buffer, bli_buffer); 
      
	if(SetProgress(70, "Localizando cruce cero...")) return false;
/* Perform edge detection using bli and gradient thresholding */
	LocateZeroCrossings (buffer, smoothed_buffer, bli_buffer);
      
	Free_f2d(smoothed_buffer); 
	
	if(SetProgress(90, "Thresholding...")) return false;
	ThresholdEdges(buffer);
	for (i = 0; i < GetRows(); i++)
	  for (j=0; j < GetCols(); j++)
	    if (m_res[i][j] > 0) m_res[i][j] = 0;
	    else m_res[i][j] = 255;
	
	Free_f2d(buffer); 
	return true;
}

/*	Recursive filter realization of the ISEF 
	(Shen and Castan CVIGP March 1992)	 */
void CShen::ComputeISEF (float **x, float **y)
{
	float **A, **B;
   
	A = f2d(); /* store causal component */
	B = f2d(); /* store anti-causal component */
   
/* first apply the filter in the vertical direcion (to the rows) */
	ApplyISEFvertical (x, y, A, B);
   
/* now apply the filter in the horizontal direction (to the columns) and */
/* apply this filter to the results of the previous one */
	ApplyISEFhorizontal (y, y, A, B);
   
   /* free up the memory */
	Free_f2d(B);
	Free_f2d(A);
}

void CShen::ApplyISEFvertical (float **x, float **y, float **A, float **B)
{
	register int row, col;
	float b1, b2;
   
	b1 = (float)((1.0 - m_b)/(1.0 + m_b));
	b2 = (float)(m_b * b1);
   
/* compute boundary conditions */
	for (col=0; col<GetCols(); col++)
	{

/* boundary exists for 1st and last column */
	   A[0][col] = b1 * x[0][col];	
	   B[GetRows()-1][col] = b2 * x[GetRows()-1][col];
	}
   
/* compute causal component */
	for (row = 1; row < GetRows(); row++)
	  for (col = 0; col < GetCols(); col++)
	    A[row][col] = (float) (b1 * x[row][col] + m_b * A[row-1][col]);

/* compute anti-causal component */
	for (row = GetRows() - 2; row >= 0; row--)
	  for (col = 0; col < GetCols(); col++)
	    B[row][col] = (float) (b2 * x[row][col] + m_b * B[row+1][col]);

/* boundary case for computing output of first filter */
	for (col = 0; col < GetCols() - 1; col++)
	  y[GetRows()-1][col] = A[GetRows()-1][col]; 

/* now compute the output of the first filter and store in y */
/* this is the sum of the causal and anti-causal components */
	for (row = 0; row < GetRows() - 2; row++)
	  for (col = 0; col < GetCols()-1; col++)
	    y[row][col] = A[row][col] + B[row+1][col];
}  


void CShen::ApplyISEFhorizontal (float **x, float **y, float **A, float **B)
{
	register int row, col;
	float b1, b2;
   
	b1 = (float) ((1.0 - m_b)/(1.0 + m_b));
	b2 = (float) (m_b * b1);
   
/* compute boundary conditions */
	for (row = 0; row < GetRows(); row++)
	{
	   A[row][0] = b1 * x[row][0];
	   B[row][GetCols()-1] = b2 * x[row][GetCols()-1];
	}

/* compute causal component */
	for (col = 1; col < GetCols(); col++)
	  for (row=0; row< GetRows(); row++)
	    A[row][col] = (float) (b1 * x[row][col] + m_b * A[row][col-1]);

/* compute anti-causal component */
	for (col = GetCols() - 2; col>=0; col--)
	  for (row = 0; row < GetRows();row++)
	    B[row][col] = (float) (b2 * x[row][col] + m_b * B[row][col+1]);

/* boundary case for computing output of first filter */     
	for (row = 0; row < GetRows(); row++)
	  y[row][GetCols()-1] = A[row][GetCols()-1];

/* now compute the output of the second filter and store in y */
/* this is the sum of the causal and anti-causal components */
	for (row = 0; row < GetRows(); row++)
	  for (col = 0; col < GetCols() - 1; col++)
	    y[row][col] = A[row][col] + B[row][col + 1];
}  

/* compute the band-limited laplacian of the input image */
void CShen::ComputeBli (float **buff1, float **buff2, CImage& bli_buffer)
{
	register int row, col;
   
	for (row = 0; row < GetRows(); row++)
	  for (col = 0; col < GetCols(); col++)
	    bli_buffer[row][col] = 0;
   
/* The bli is computed by taking the difference between the smoothed image */
/* and the original image.  In Shen and Castan's paper this is shown to */
/* approximate the band-limited laplacian of the image.  The bli is then */
/* made by setting all values in the bli to 1 where the laplacian is */
/* positive and 0 otherwise. 						*/
	for (row=0; row < GetRows(); row++)
	  for (col=0; col < GetCols(); col++)
	  {
            if (row < OUTLINE || row >= GetRows()-OUTLINE ||
                  col < OUTLINE || col >= GetCols()-OUTLINE) continue;
	    bli_buffer[row][col] = 
		((buff1[row][col] - buff2[row][col]) > 0.0);
	  }
}

void CShen::LocateZeroCrossings (float **orig, float **smoothed, CImage& bli)
{
	register int row, col;
   
	for (row = 0; row < GetRows(); row++)
	{
	   for (col=0; col<GetCols(); col++)
	   {

/* ignore pixels around the boundary of the image */
	   if (row < OUTLINE || row >= GetRows() - OUTLINE ||
		 col < OUTLINE || col >= GetCols() - OUTLINE)
	   {
	     orig[row][col] = 0.0;
	   }

 /* next check if pixel is a zero-crossing of the laplacian  */
	   else if (IsCandidateEdge (bli, smoothed, row, col))
	   {

/* now do gradient thresholding  */
	      float grad = ComputeAdaptiveGradient (bli, smoothed, row, col);
	      orig[row][col] = grad;
	   }
	   else  orig[row][col] = 0.0;		    
	  }
	}
}

void CShen::ThresholdEdges (float **in)
{
	register int i, j;
	double low_thresh = m_low_thresh, high_thresh = m_high_thresh;
   
	m_lap = in;
   
	EstimateThresh(&low_thresh, &high_thresh);
	if (!m_do_hysteresis)
	  low_thresh = high_thresh;

	for (i=0; i < GetRows(); i++)
	  for (j=0; j < GetCols(); j++)
	    m_res[i][j] = 0;
   
	for (i=0; i < GetRows(); i++)
	  for (j=0; j < GetCols(); j++)
	  {
            if (i < OUTLINE || i >= GetRows()-OUTLINE ||
                  j < OUTLINE || j >= GetCols()-OUTLINE) continue;

/* only check a contour if it is above high_thresh */
	    if ((m_lap[i][j]) > high_thresh) 

/* mark all connected points above low thresh */
	      MarkConnected(i,j,0, low_thresh);	
	  }
   
	for (i = 0; i < GetRows(); i++)	/* erase all points which were 255 */
	  for (j = 0; j < GetCols(); j++)
	    if (m_res[i][j] == 255) m_res[i][j] = 0;
}

/*	return true if it marked something */ 
int CShen::MarkConnected (int i, int j, int level, double low_thresh)
{
	 int notChainEnd;
    
   /* stop if you go off the edge of the image */
	if (i >= GetRows() || i < 0 || j >= GetCols() || j < 0) return 0;
   
   /* stop if the point has already been visited */
	if (m_res[i][j] != 0) return 0;	
   
   /* stop when you hit an image boundary */
	if (m_lap[i][j] == 0.0) return 0;
   
	if ((m_lap[i][j]) > low_thresh)
	{
	   m_res[i][j] = 1;
	}
	else
	{
	   m_res[i][j] = 255;
	}
   
	notChainEnd =0;
    
	notChainEnd |= MarkConnected(i  ,j+1, level+1, low_thresh);
	notChainEnd |= MarkConnected(i  ,j-1, level+1, low_thresh);
	notChainEnd |= MarkConnected(i+1,j+1, level+1, low_thresh);
	notChainEnd |= MarkConnected(i+1,j  , level+1, low_thresh);
	notChainEnd |= MarkConnected(i+1,j-1, level+1, low_thresh);
	notChainEnd |= MarkConnected(i-1,j-1, level+1, low_thresh);
	notChainEnd |= MarkConnected(i-1,j  , level+1, low_thresh);
	notChainEnd |= MarkConnected(i-1,j+1, level+1, low_thresh);

	if (notChainEnd && ( level > 0 ) )
	{
	/* do some contour thinning */
	  if ( m_thinFactor > 0 )
	  if ( (level % m_thinFactor) != 0  )
	  {
	    /* delete this point */  
	    m_res[i][j] = 255;
	  }
	}
    
	return 1;
}

/* finds zero-crossings in laplacian (buff)  orig is the smoothed image */
int CShen::IsCandidateEdge (CImage& buff, float **orig, int row, int col)
{
/* test for zero-crossings of laplacian then make sure that zero-crossing */
/* sign correspondence principle is satisfied.  i.e. a positive z-c must */
/* have a positive 1st derivative where positive z-c means the 2nd deriv */
/* goes from positive to negative as we pass through the step edge */
   
	if (buff[row][col] == 1 && buff[row+1][col] == 0) /* positive z-c */
	{ 
	   if (orig[row+1][col] - orig[row-1][col] > 0) return 1;
	   else return 0;
	}
	else if (buff[row][col] == 1 && buff[row][col+1] == 0 ) /* positive z-c */
	{
	   if (orig[row][col+1] - orig[row][col-1] > 0) return 1;
	   else return 0;
	}
	else if ( buff[row][col] == 1 && buff[row-1][col] == 0) /* negative z-c */
	{
	   if (orig[row+1][col] - orig[row-1][col] < 0) return 1;
	   else return 0;
	}
	else if (buff[row][col] == 1 && buff[row][col-1] == 0 ) /* negative z-c */
	{
	   if (orig[row][col+1] - orig[row][col-1] < 0) return 1;
	   else return 0;
	}
	else			/* not a z-c */
	  return 0;
}

float CShen::ComputeAdaptiveGradient (CImage& BLI_buffer, float **orig_buffer, 
				 int row, int col)
{
	register int i, j;
	float sum_on, sum_off;
	float avg_on, avg_off;
	int num_on, num_off;
   
	sum_on = sum_off = 0.0;
	num_on = num_off = 0;
   
	for (i= (-m_window_size/2); i<=(m_window_size/2); i++)
	{
	   for (j=(-m_window_size/2); j<=(m_window_size/2); j++)
	   {
	     if (BLI_buffer[row+i][col+j])
	     {
	        sum_on += orig_buffer[row+i][col+j];
	        num_on++;
	     }
	     else
	     {
	        sum_off += orig_buffer[row+i][col+j];
	        num_off++;
	     }
	   }
	}
   
	if (sum_off) avg_off = sum_off / num_off;
	else avg_off = 0;
   
	if (sum_on) avg_on = sum_on / num_on;
	else avg_on = 0;
   
	return (avg_off - avg_on);
}

void CShen::EstimateThresh (double *low, double *hi)
{
	float vmax, vmin, scale, x;
	int i,j,k, hist[256], count;

/* Build a histogram of the Laplacian image. */
	vmin = vmax = (float) fabs((float)(m_lap[20][20]));
	for (i=0; i<GetRows(); i++)
	  for (j=0; j<GetCols(); j++)
	  {
            if (i < OUTLINE || i >= GetRows()-OUTLINE ||
                  j < OUTLINE || j >= GetCols()-OUTLINE) continue;
	    x = m_lap[i][j];
	    if (vmin > x) vmin = x;
	    if (vmax < x) vmax = x;
	  }
	for (k=0; k<256; k++) hist[k] = 0;

	scale = (float) (256.0/(vmax-vmin + 1));

	for (i=0; i < GetRows(); i++)
	  for (j=0; j < GetCols(); j++)
	  {
            if (i<OUTLINE || i >= GetRows()-OUTLINE ||
                  j<OUTLINE || j >= GetCols()-OUTLINE) continue;
	    x = m_lap[i][j];
	    k = (int)((x - vmin)*scale);
	    hist[k] += 1;
	  }

/* The high threshold should be > 80 or 90% of the pixels */
	k = 255;
	j = (int)(m_ratio * GetRows() * GetCols());
	count = hist[255];
	while (count < j)
	{
	  k--;
	  if (k<0) break;
	  count += hist[k];
	}
	*hi = (double)k/scale   + vmin ;
	*low = (*hi)/2;
}
